import json
import sys
from enum import Enum

from attrs import define
from datetime import datetime
from decimal import Decimal

from src.seriattrs import DbClass
from src.seriattrs import DbClassLiteral


def test_serialize_literal():
    class Color(Enum):
        RED = 'red'

    @define
    class Bar(DbClassLiteral):
        dictionary: dict
        date: datetime
        decimal: Decimal
        color: Color

    @define
    class Foo(DbClass):
        dictionary: dict
        date: datetime
        decimal: Decimal
        bar: Bar
        color: Color

    foo = Foo({}, datetime.now(), Decimal(1), Bar({}, datetime.now(), Decimal(1), Color.RED), Color.RED)
    serialized = foo.serialize()
    try:
        json.dump(serialized, sys.stdout)
    except:
        assert False
    deserialized = Foo.deserialize(serialized)
    assert deserialized == foo


def test_serialize():
    class Color(Enum):
        RED = 'red'

    @define
    class Bar(DbClass):
        dictionary: dict
        date: datetime
        decimal: Decimal
        color: Color

    @define
    class Foo(DbClass):
        dictionary: dict
        date: datetime
        decimal: Decimal
        bar: Bar
        bars: tuple[Bar, Bar]
        bars_dict: dict[Bar, Bar]
        color: Color

    foo = Foo({}, datetime.now(), Decimal(1), Bar({}, datetime.now(), Decimal(1), Color.RED), (Bar({}, datetime.now(), Decimal(1), Color.RED), Bar({}, datetime.now(), Decimal(1), Color.RED),), {Bar({}, datetime.now(), Decimal(1), Color.RED): Bar({}, datetime.now(), Decimal(1), Color.RED)}, Color.RED)
    serialized = foo.serialize()
    foo.bar = foo.bar._id
    try:
        json.dump(serialized, sys.stdout)
    except:
        assert False
    deserialized = Foo.deserialize(serialized)
    assert deserialized == foo



from abc import ABC
from typing import get_args, Self

from attr import fields
from attrs import define

from .DbClass import DbClass
from .._db_attrs_converter import _db_attrs_converter


@define
class DbClassLiteral(ABC, DbClass):
    """Database Class that implements serialize and deserialize methods.
    Each object of a class is provided with int64 number as an id.
    DbClassLiteral is an abstract class and must be extended in order to create new instances.

    :var _id: int64 value provided by default to each instance of a DbClass. Can't be initialized but can be changed after initialization.
    """


def _handle_new_db(value, db_class_type):
    if isinstance(value, int):
        return value
    if db_class_type.__name__ not in type(db_class_type).temp_instances:
        type(db_class_type).temp_instances[db_class_type.__name__] = {}
    if value["_id"] in type(db_class_type).temp_instances[db_class_type.__name__]:
        return type(db_class_type).temp_instances[db_class_type.__name__][value["_id"]]
    else:
        if db_class_type not in type(db_class_type)._created_types.values():
            db_class_type = next(class_ for class_ in type(db_class_type)._created_types.values() if class_.__name__ == db_class_type.__name__)
        id_value = value.pop("_id")
        new_instance = db_class_type(**value)
        new_instance._id = id_value
        value["_id"] = id_value
        type(db_class_type).temp_instances[db_class_type.__name__][id_value] = new_instance
    for f in fields(db_class_type):
        types = list(get_args(f.type))
        if isinstance(f.type, type):
            types.append(f.type)
        references = tuple(field_type for field_type in types if isinstance(field_type, str))
        if references:
            raise ValueError("Data can't be deserialized as field {} in {} has a forward references {}. Consider making classes {} DbClasses and initializing them.".format(f.name, db_class_type.__name__, references, references))
        try:
            if any(issubclass(db_class_type if field_type == Self else field_type, DbClassLiteral) for field_type in types):
                setattr(
                    new_instance,
                    f.name,
                    _db_attrs_converter.structure(value[f.name], f.type),
                )
        except Exception as e:
            pass
    return new_instance


_db_attrs_converter.register_structure_hook_func(
    lambda t: issubclass(t, DbClass), _handle_new_db
)



import json
from dataclasses import is_dataclass, asdict
from typing import Any
from ...JsonEncoder import Encoder


class DataclassEncoder(Encoder):
    @staticmethod
    def is_valid(element: Any) -> bool:
        return is_dataclass(element)

    @staticmethod
    def encode(element) -> dict:
        from ..DefaultJsonEncoder import DefaultJsonEncoder

        return json.loads(json.dumps(asdict(element), cls=DefaultJsonEncoder))



from typing import Any, Mapping

from attr import has
from attr._make import fields

from ...JsonEncoder import Encoder


class AttrsEncoder(Encoder):
    @staticmethod
    def is_valid(element: Any) -> bool:
        return has(element)

    @staticmethod
    def encode(element) -> dict:
        from ..DefaultJsonEncoder import DefaultJsonEncoder
        from ... import DbClassLiteral, DbClass

        if isinstance(element, DbClassLiteral) or not isinstance(element, DbClass):
            memory = {}
            dictionary = asdict(element, memory=memory)
            fill_memory_gaps(memory, memory)
            return DefaultJsonEncoder.serialize_values(dictionary)
        return element._id


def fill_memory_gaps(memory_item, memory, short_term_memory = None):
    if short_term_memory is None:
        short_term_memory = set()
    if id(memory_item) in short_term_memory:
        return
    else:
        short_term_memory.add(id(memory_item))
    if isinstance(memory_item, Mapping):
        for key, value in tuple(memory_item.items()):
            if isinstance(value, _Id):
                memory_item[key] = memory[memory_item[key]]
            else:
                fill_memory_gaps(value, memory, short_term_memory)
    elif isinstance(memory_item, list | set):
        for index, item in enumerate(memory_item):
            if isinstance(item, _Id):
                memory_item[index] = memory[item]
            else:
                fill_memory_gaps(item, memory, short_term_memory)
    elif isinstance(memory_item, _Id):
        raise ValueError

class _Id(int):
    pass



def asdict(
    inst,
    recurse=True,
    filter=None,
    dict_factory=dict,
    retain_collection_types=False,
    value_serializer=None,
    memory=None,
):
    """
    Return the *attrs* attribute values of *inst* as a dict.

    Optionally recurse into other *attrs*-decorated classes.

    :param inst: Instance of an *attrs*-decorated class.
    :param bool recurse: Recurse into classes that are also
        *attrs*-decorated.
    :param callable filter: A callable whose return code determines whether an
        attribute or element is included (``True``) or dropped (``False``).  Is
        called with the `attrs.Attribute` as the first argument and the
        value as the second argument.
    :param callable dict_factory: A callable to produce dictionaries from.  For
        example, to produce ordered dictionaries instead of normal Python
        dictionaries, pass in ``collections.OrderedDict``.
    :param bool retain_collection_types: Do not convert to ``list`` when
        encountering an attribute whose type is ``tuple`` or ``set``.  Only
        meaningful if ``recurse`` is ``True``.
    :param Optional[callable] value_serializer: A hook that is called for every
        attribute or dict key/value.  It receives the current instance, field
        and value and must return the (updated) value.  The hook is run *after*
        the optional *filter* has been applied.

    :rtype: return type of *dict_factory*

    :raise attrs.exceptions.NotAnAttrsClassError: If *cls* is not an *attrs*
        class.

    ..  versionadded:: 16.0.0 *dict_factory*
    ..  versionadded:: 16.1.0 *retain_collection_types*
    ..  versionadded:: 20.3.0 *value_serializer*
    ..  versionadded:: 21.3.0 If a dict has a collection for a key, it is
        serialized as a tuple.
    """
    if memory is None:
        memory = {}
    if id(inst) not in memory:
        memory[id(inst)] = None
    elif memory[id(inst)] is None:
        memory[id(inst)] = _Id(id(inst))
        return memory[id(inst)]
    else:
        return memory[id(inst)]
    attrs = fields(inst.__class__)
    rv = dict_factory()
    for a in attrs:
        v = getattr(inst, a.name)
        if filter is not None and not filter(a, v):
            continue

        if value_serializer is not None:
            v = value_serializer(inst, a, v)

        if recurse is True:
            if has(v.__class__):
                rv[a.name] = asdict(
                    v,
                    recurse=True,
                    filter=filter,
                    dict_factory=dict_factory,
                    retain_collection_types=retain_collection_types,
                    value_serializer=value_serializer,
                    memory=memory,
                )
            elif isinstance(v, (tuple, list, set, frozenset)):
                cf = v.__class__ if retain_collection_types is True else list
                rv[a.name] = cf(
                    [
                        _asdict_anything(
                            i,
                            is_key=False,
                            filter=filter,
                            dict_factory=dict_factory,
                            retain_collection_types=retain_collection_types,
                            value_serializer=value_serializer,
                            memory=memory,
                        )
                        for i in v
                    ]
                )
            elif isinstance(v, dict):
                df = dict_factory
                rv[a.name] = df(
                    (
                        _asdict_anything(
                            kk,
                            is_key=True,
                            filter=filter,
                            dict_factory=df,
                            retain_collection_types=retain_collection_types,
                            value_serializer=value_serializer,
                            memory=memory,
                        ),
                        _asdict_anything(
                            vv,
                            is_key=False,
                            filter=filter,
                            dict_factory=df,
                            retain_collection_types=retain_collection_types,
                            value_serializer=value_serializer,
                            memory=memory,
                        ),
                    )
                    for kk, vv in v.items()
                )
            else:
                rv[a.name] = v
        else:
            rv[a.name] = v
    memory[id(inst)] = rv
    return rv


def _asdict_anything(
    val,
    is_key,
    filter,
    dict_factory,
    retain_collection_types,
    value_serializer,
    memory,
):
    """
    ``asdict`` only works on attrs instances, this works on anything.
    """
    if id(val) in memory:
        return memory[id(val)]
    if getattr(val.__class__, "__attrs_attrs__", None) is not None:
        # Attrs class.
        rv = asdict(
            val,
            recurse=True,
            filter=filter,
            dict_factory=dict_factory,
            retain_collection_types=retain_collection_types,
            value_serializer=value_serializer,
            memory=memory,
        )
    elif isinstance(val, (tuple, list, set, frozenset)):
        if retain_collection_types is True:
            cf = val.__class__
        elif is_key:
            cf = tuple
        else:
            cf = list

        rv = cf(
            [
                _asdict_anything(
                    i,
                    is_key=False,
                    filter=filter,
                    dict_factory=dict_factory,
                    retain_collection_types=retain_collection_types,
                    value_serializer=value_serializer,
                    memory=memory,
                )
                for i in val
            ]
        )
    elif isinstance(val, dict):
        df = dict_factory
        rv = df(
            (
                _asdict_anything(
                    kk,
                    is_key=True,
                    filter=filter,
                    dict_factory=df,
                    retain_collection_types=retain_collection_types,
                    value_serializer=value_serializer,
                    memory=memory,
                ),
                _asdict_anything(
                    vv,
                    is_key=False,
                    filter=filter,
                    dict_factory=df,
                    retain_collection_types=retain_collection_types,
                    value_serializer=value_serializer,
                    memory=memory,
                ),
            )
            for kk, vv in val.items()
        )
    else:
        rv = val
        if value_serializer is not None:
            rv = value_serializer(None, None, rv)
    memory[id(val)] = rv
    return rv



__all__ = [
    "Decoder",
    "Encoder",
    "DefaultJsonEncoder",
]

from .Decoder import Decoder
from .Encoder import Encoder
from .Encoders import *
from .Decoders import *
from .DefaultJsonEncoder import DefaultJsonEncoder



import random
from typing import Any, Self

from attr import define, fields, field

from .DbClassCreator import DbClassCreator
from ..JsonEncoder import Decoder, DefaultJsonEncoder
from .._db_attrs_converter import _db_attrs_converter


@define
class DbClass(metaclass=DbClassCreator):
    """Database Class that implements serialize and deserialize methods.
    Each object of a class is provided with int64 number as an id.
    Whenever field contains a DbClass instance _id field on DbClass is used instead. If an actual class is to be used use DbClassLiteral as a base.
    DbClass is an abstract class and must be extended in order to create new instances.

    :var _id: int64 value provided by default to each instance of a DbClass. Can't be initialized but can be changed after initialization.
    """
    _id = field(init=False, type=Any, factory=lambda: random.randint(-2 ** 63, 2 ** 63 - 1))

    def __attrs_post_init__(self):
        if hasattr(self, 'id'):
            self._id = self.id
        self._decode()

    def serialize(self) -> dict:
        from .DbClassLiteral import DbClassLiteral

        return DefaultJsonEncoder.serialize_values(
            dict(
                (
                    f.name,
                    getattr(self, f.name)._id
                    if isinstance(getattr(self, f.name), DbClass)
                       and not isinstance(getattr(self, f.name), DbClassLiteral)
                    else getattr(self, f.name),
                )
                for f in fields(type(self))
            ),
        )

    @classmethod
    def deserialize(cls, dictionary: dict) -> Self:
        type(cls).temp_instances = {}
        deserialized = _db_attrs_converter.structure(dictionary, cls)
        type(cls).temp_instances = {}
        return deserialized

    def _decode(self):
        for f in fields(type(self)):
            for decoder in Decoder.__subclasses__():
                if decoder.is_valid(f.type):
                    setattr(self, f.name, decoder.decode(getattr(self, f.name), f.type))
                    break



from enum import Enum
from typing import Any, TypeVar, Type

from ...JsonEncoder import Decoder

T = TypeVar('T', bound=Enum)


class EnumDecoder(Decoder):
    @staticmethod
    def is_valid(element: Any) -> bool:
        if not isinstance(element, type):
            return False
        return issubclass(element, Enum)

    @staticmethod
    def decode(element: Any, enum: Type[T]) -> T:
        return enum(element)



from typing import Any

from ...JsonEncoder import Decoder


class BytesDecoder(Decoder):
    @staticmethod
    def is_valid(element: Any) -> bool:
        return isinstance(element, str)

    @staticmethod
    def decode(element: str, _) -> bytes:
        return element.encode()



from json import JSONEncoder
from typing import Mapping

from .Encoder import Encoder


class DefaultJsonEncoder(JSONEncoder):
    @classmethod
    def default(cls, obj, strict: bool = True):
        for encoder in Encoder.__subclasses__():
            if encoder.is_valid(obj):
                return encoder.encode(obj)
        else:
            if strict:
                raise ValueError(obj, "can't be encoded")
            return obj

    @classmethod
    def serialize_values(cls, values, short_term_memory=None):
        if short_term_memory is None:
            short_term_memory = dict()
        if id(values) in short_term_memory:
            return short_term_memory[id(values)]
        else:
            short_term_memory[id(values)] = values
        if isinstance(values, Mapping):
            short_term_memory[id(values)] = dict(
                (key, cls.serialize_values(value, short_term_memory)) for key, value in values.items())
            return short_term_memory[id(values)]
        elif isinstance(values, list | set):
            short_term_memory[id(values)] = list(cls.serialize_values(item, short_term_memory) for item in values)
            return short_term_memory[id(values)]
        else:
            short_term_memory[id(values)] = cls._serialize_value(values)
            return short_term_memory[id(values)]

    @classmethod
    def _serialize_value(cls, obj):
        return DefaultJsonEncoder().default(obj, strict=False)



from abc import abstractmethod
from typing import Any
from ..JsonEncoder.JsonOperator import JSONOperator


class Encoder(JSONOperator):
    @staticmethod
    @abstractmethod
    def is_valid(element: Any) -> bool:
        pass

    @staticmethod
    @abstractmethod
    def encode(element: Any) -> str:
        pass



from pathlib import Path
from typing import Any

from ...JsonEncoder import Decoder


class PathDecoder(Decoder):
    @staticmethod
    def is_valid(element: Any) -> bool:
        return isinstance(element, str)

    @staticmethod
    def decode(element: str, _) -> Path:
        return Path(element)



from functools import partial

from attr import field, validators


def _check_int(instance, attribute, value, bits: int):
    if value < -(2 ** (bits - 1)):
        raise ValueError(f"{value=} must be more than {-(2 ** (bits - 1))}")
    if value > 2 ** (bits - 1) - 1:
        raise ValueError(f"{value=} must be less than {2 ** (bits - 1) - 1}")


def _check_uint(instance, attribute, value, bits: int):
    if value < 0:
        raise ValueError(f"{value=} must be positive")
    if value > 2**bits - 1:
        raise ValueError(f"{value=} must be less than {2 ** bits - 1}")

#: Field which can be used to store int8 values
int8 = partial(
    field, validator=[validators.instance_of(int), partial(_check_int, bits=8)]
)
#: Field which can be used to store int16 values
int16 = partial(
    field, validator=[validators.instance_of(int), partial(_check_int, bits=16)]
)
#: Field which can be used to store int32 values
int32 = partial(
    field, validator=[validators.instance_of(int), partial(_check_int, bits=32)]
)
#: Field which can be used to store int64 values
int64 = partial(
    field, validator=[validators.instance_of(int), partial(_check_int, bits=64)]
)
#: Field which can be used to store uint8 values
uint8 = partial(
    field, validator=[validators.instance_of(int), partial(_check_uint, bits=8)]
)
#: Field which can be used to store uint16 values
uint16 = partial(
    field, validator=[validators.instance_of(int), partial(_check_uint, bits=16)]
)
#: Field which can be used to store uint32 values
uint32 = partial(
    field, validator=[validators.instance_of(int), partial(_check_uint, bits=32)]
)
#: Field which can be used to store uint64 values
uint64 = partial(
    field, validator=[validators.instance_of(int), partial(_check_uint, bits=64)]
)



import re
import typing
from abc import ABCMeta
from collections import defaultdict
from inspect import signature
from types import GenericAlias
from typing import Any, get_args, ForwardRef

from attr import fields

_ = typing


class DbClassCreator(ABCMeta):
    """Metaclass that converts Forward references to class instances.

    @define
    class Bar(DbClass):
        foo: "Foo" => Foo

    @define
    class Foo(DbClass):
        bar: "Bar" => Bar
    """
    _type_checked_fields = defaultdict(list)
    _created_types = {}

    def __new__(cls, name: str, bases: tuple, namespace: dict[str, Any]):
        new_class = super().__new__(cls, name, bases, namespace)
        for field_name, field_type in new_class.__annotations__.items():
            forward_refs = cls._get_forward_refs(field_type)
            for forward_ref in forward_refs:
                if forward_ref in cls._created_types:
                    new_class.__annotations__[field_name] = cls._update_hint(forward_ref,
                                                                             new_class.__annotations__[field_name])
                else:
                    cls._type_checked_fields[forward_ref].append(new_class)
        cls._created_types[ForwardRef(name)] = new_class
        for db_class in cls._type_checked_fields.get(ForwardRef(name), []):
            attrs_fields = []
            for f in fields(db_class):
                forward_refs = cls._get_forward_refs(f.type)
                current_string_type = f.type
                for forward_ref in forward_refs:
                    if forward_ref not in cls._created_types:
                        continue
                    current_string_type = cls._update_hint(forward_ref, current_string_type)
                field_value = dict((arg, getattr(f, arg)) for arg in signature(type(f)).parameters if hasattr(f, arg))
                field_value['type'] = current_string_type
                field_value['cmp'] = None
                new_field = type(f)(**field_value)
                attrs_fields.append(new_field)
            db_class.__attrs_attrs__ = type(db_class.__attrs_attrs__)(attrs_fields)
        return new_class

    @classmethod
    def _get_forward_refs(cls, field_type):
        forward_refs = list(ForwardRef(ref) if isinstance(ref, str) else ref for ref in get_args(field_type) if
                            isinstance(ref, (str, ForwardRef)))
        if isinstance(field_type, str):
            forward_refs.append(ForwardRef(field_type))
        return forward_refs

    @classmethod
    def _update_hint(cls, forward_ref: ForwardRef, current_string_type: GenericAlias) -> GenericAlias:
        forward_ref_name = cls._created_types[forward_ref].__name__
        return eval(
            re.sub(r'(?:ForwardRef\(|)\'' + forward_ref_name + r'\'\)?', forward_ref_name, str(current_string_type)),
            globals(), {forward_ref_name: cls._created_types[forward_ref]}
        )



from typing import Any

from ...JsonEncoder.Encoder import Encoder


class BytesEncoder(Encoder):
    @staticmethod
    def is_valid(element: Any) -> bool:
        return isinstance(element, bytes)

    @staticmethod
    def encode(element: bytes) -> str:
        return element.decode()



from pathlib import Path
from typing import Any

from ...JsonEncoder.Encoder import Encoder


class PathEncoder(Encoder):
    @staticmethod
    def is_valid(element: Any) -> bool:
        return isinstance(element, Path)

    @staticmethod
    def encode(element: Path) -> str:
        return str(element)
